﻿Imports KernCADnet

Public Class SplineMods

    Enum State
        eStateIdle
        eStatePickingPoint
        eStatePickingMoveToPoint
        eStateReadyToMove
        eStateDone
    End Enum

    Dim m_state As State = State.eStateIdle
    Dim m_iGener As IObjectGenerator_DG     ' Generic constructor
    Dim m_iModel As IModel_DG
    Dim m_iStdShape As IStdShape_DG
    Dim m_iPick As IPick
    Dim m_iEntity As IEntity_DG             ' Entity which has the curve
    Dim m_iSplineCurve As IBSplineCurve_DG
    Dim m_pt As DIPoint                     ' The picked point to be modified
    Dim m_ptNew As DIPoint                  ' New modified position of the picked point
    Dim m_iArrPnt As IPointArray_DG         ' Definition points for the spline

    Private Sub OnFormLoad(sender As Object, e As EventArgs) Handles MyBase.Load
        m_iModel = m_dgk.GetModel()
        m_iGener = m_iModel
        m_iStdShape = m_iModel

        CreateSpline()

        ' Add polygon of poles to the scene
        DisplayPolesPolygon(m_iSplineCurve)
        DisplayInterpolationPoints(m_iArrPnt)       ' Show the points

        ZoomExtents(2.5)
        UpdatePage()
    End Sub

    Sub ZoomExtents(zoom As Double)

        Dim iView As IView_DG = m_dgk.GetView()
        ' Fit into the window
        iView.Reset(True, True)
        ' Zoom out
        iView.Zoom(zoom)
    End Sub

    Sub CreateSpline()
        If m_iArrPnt Is Nothing Then m_iArrPnt = m_iGener.Create("PointArray_DG")

        m_iArrPnt.SetCount(5)
        Dim pt As PointDg = New PointDg()
        pt.Set(0, 0, 10) : m_iArrPnt.SetPoint(0, pt)
        pt.Set(10, 20, 20) : m_iArrPnt.SetPoint(1, pt)
        pt.Set(20, 30, 30) : m_iArrPnt.SetPoint(2, pt)
        pt.Set(40, 30, 40) : m_iArrPnt.SetPoint(3, pt)
        pt.Set(50, 50, 50) : m_iArrPnt.SetPoint(4, pt)

        m_iSplineCurve = m_iGener.Create("BSplineCurve_DG")

        m_iSplineCurve.InitFromPoints(m_iArrPnt, 3, 8, Continuity_DG.eCont_C2, 0.001)

        Dim iShape As IBRepShape_DG = m_iGener.Create1("BRepEdge_DG", m_iSplineCurve)

        ' Add shape to the displayed model
        Dim entity As IEntity_DG = m_iModel.AddBRepShape(iShape)

        entity.SetColor(1, 0, 0)        'Red
    End Sub

    Sub DisplayPolesPolygon(iCurve As IBSplineCurve_DG)
        Dim iArrPnt As IPointArray_DG = iCurve.GetPoles()

        Dim n As Integer = iArrPnt.GetCount()
        ' Create segment strip geometry and entity
        Dim iEntityPolygon As IEntity_DG = m_iModel.AddNewEntity1("SegmentStrip")
        iEntityPolygon.SetColor(0, 0, 1)    'Blue in rgb

        ' Copy points
        Dim iPointSet As ISectionPointSet = iEntityPolygon
        Dim iArray As IArray2 = iPointSet
        iArray.SetCount(n)
        Dim points As IMatrixData = iPointSet.GetPoints()
        For k As Integer = 0 To n - 1
            Dim pt As Point_DG = iArrPnt.GetPoint(k)
            points.Set(k, 0, pt.x(0))
            points.Set(k, 1, pt.x(1))
            points.Set(k, 2, pt.x(2))
        Next k
    End Sub

    Sub AddSpherePoint(pt As Point_DG, color As String)
        Dim pEnt As IEntity_DG = m_iStdShape.Sphere(0.5)
        Dim iFrame As IFrame_DG = pEnt.GetLocation()
        iFrame.SetOrigin(pt)
        If color = "Yellow" Then
            pEnt.SetColor(1.0F, 1.0F, 0)
        ElseIf color = "Red" Then
            pEnt.SetColor(1.0F, 0.0F, 0)
        ElseIf color = "Blue" Then
            pEnt.SetColor(0.0F, 0.0F, 1.0F)
        End If
        m_iModel.AddEntity(pEnt)
    End Sub

    Sub DisplayInterpolationPoints(iArrPnt As IPointArray_DG)

        Dim n As Integer = iArrPnt.GetCount()
        ' Create point set geometry and entity
        Dim iEntityPoints As IEntity_DG = m_iModel.AddNewEntity1("PointArray")

        ' Copy points
        Dim iPointSet As ISectionPointSet = iEntityPoints
        Dim iArray As IArray2 = iPointSet
        iArray.SetCount(n)
        Dim points As IMatrixData = iPointSet.GetPoints()

        For k As Integer = 0 To n - 1
            Dim pt As Point_DG = iArrPnt.GetPoint(k)
            points.Set(k, 0, pt.x(0))
            points.Set(k, 1, pt.x(1))
            points.Set(k, 2, pt.x(2))

            AddSpherePoint(pt, "Yellow")
        Next k

        iPointSet.PointSize = 5     ' In pixels
    End Sub

    Sub ModifySpline()
        Dim iCurve As ICurve_DG = m_iSplineCurve

        Dim pt As Point_DG = New PointDg(m_pt.x(0), m_pt.x(1), m_pt.x(2))
        Dim ptNearest As Point_DG
        Dim u As Double = iCurve.GetNearestPoint(pt, ptNearest)

        Dim i1 As Integer = FindPoleSegment(iCurve, u)

        Dim ptNew As PointDg = New PointDg(m_ptNew.x(0), m_ptNew.x(1), m_ptNew.x(2))

        Dim k0, k1 As Integer
        m_iSplineCurve.ModifyPoint(u, ptNew, i1 - 1, i1, k0, k1)

        ' Set modified flag. Curve is too low-level attribute to know how to notify entity about the modification
        Dim iGeometry As IGeometry_DG = m_iEntity.GetGeometry()
        iGeometry.SetModified()
    End Sub

    ' Find two sequential poles adjacent to the picked point. Return index of the second
    Function FindPoleSegment(iCurve As ICurve_DG, u As Double) As Integer
        Dim n As Integer = m_iArrPnt.GetCount()

        Dim ptNearest As Point_DG
        ' Find the first definition point with parameter greater than u
        ' We also could have used projections of poles (nearest point) to the curve

        For i As Integer = 0 To n - 1
            Dim pt As PointDg = m_iArrPnt.GetPoint(i)
            Dim t As Double = iCurve.GetNearestPoint(pt, ptNearest)
            If t >= u Then Return i
        Next i
        Return -1
    End Function


    Private Sub OnStep0Begin(sender As Object, e As EventArgs) Handles buttonBegin.Click
        If m_state = State.eStateDone Then
            Close()
            Return
        End If

        m_state = State.eStatePickingPoint
        UpdatePage()
    End Sub

    Sub UpdatePage()
        buttonBegin.Enabled = m_state = State.eStateIdle Or m_state = State.eStateDone
        ButtonStep1.Enabled = m_state = State.eStatePickingPoint
        ButtonStep2.Enabled = m_state = State.eStatePickingMoveToPoint
        ButtonStep3Modify.Enabled = m_state = State.eStateReadyToMove

        If m_state = State.eStateDone Then buttonBegin.Text = "Please restart to repeat"
    End Sub

    Private Sub OnStep3Modify(sender As Object, e As EventArgs) Handles ButtonStep3Modify.Click
        If Not m_iEntity Is Nothing Then            ' Otherwise have not picked point on the first staep
            ModifySpline()
            m_dgk.UpdateView()
        End If

        m_state = State.eStateDone
        UpdatePage()
    End Sub

    Private Sub OnDgkMouseDown(sender As Object, e As AxKernCADnet._DKernCADnetEvents_MouseDownEvent) Handles m_dgk.MouseDownEvent
        If (m_state = State.eStatePickingPoint) Then

            If m_iPick Is Nothing Then
                Dim iView As IView_DG = m_dgk.GetView()
                m_iPick = iView
            End If

            Dim item As IItem = m_iPick.GetItemAt3D(e.x, e.y, m_pt)

            AddSpherePoint(New PointDg(m_pt.x(0), m_pt.x(1), m_pt.x(2)), "Red")     ' For display

            m_iEntity = item

            m_dgk.UpdateView()

            m_state = State.eStatePickingMoveToPoint
        ElseIf (m_state = State.eStatePickingMoveToPoint) Then

            Dim item As IItem = m_iPick.GetItemAt3D(e.x, e.y, m_ptNew)

            Dim ptNew As PointDg = New PointDg(m_ptNew.x(0), m_ptNew.x(1), m_ptNew.x(2))

            AddSpherePoint(ptNew, "Blue")       ' For display

            m_dgk.UpdateView()

            m_state = State.eStateReadyToMove
        End If

        UpdatePage()
    End Sub
End Class
